package com.tobee.opencv.detector;

import com.tobee.opencv.detector.tests.input.DNNOption;
import conf.CocoLabel;
import java.io.File;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import javax.swing.JOptionPane;

import org.opencv.core.Core;
import org.opencv.core.CvType;
import org.opencv.core.Mat;
import org.opencv.core.MatOfDouble;
import org.opencv.core.Point;
import org.opencv.core.Rect;
import org.opencv.core.Scalar;
import org.opencv.core.Size;
import org.opencv.dnn.Dnn;
import org.opencv.dnn.Net;
import org.opencv.highgui.HighGui;
import org.opencv.imgcodecs.Imgcodecs;
import org.opencv.imgproc.Imgproc;
import org.opencv.videoio.VideoCapture;
import org.opencv.videoio.VideoWriter;
import org.opencv.videoio.Videoio;

public class MaskRCNNDetector extends ObjectDetector {

     public MaskRCNNDetector(final DNNOption dnnOption)
     {
         super(dnnOption);
     }
     
    @Override
    public void detectObject()
    {
        if(!new File(mediaFile).exists()) {
            JOptionPane.showMessageDialog(null, String.format("File9[%s] Not Found", mediaFile), "File Error", JOptionPane.WARNING_MESSAGE);
            return;
        }
        if(!new File(modelPath).exists() || !new File(modelConfiguration).exists()) {
            JOptionPane.showMessageDialog(null, String.format("Can not open Model[%s] \n config[%s]", 
                    modelPath,modelConfiguration), "Model Error", JOptionPane.WARNING_MESSAGE);
            return;
        }
        
        Net dnnNet = readDNNetByModelName();
        if(dnnNet == null)
        {
            JOptionPane.showMessageDialog(null,"Failed to configure the network.", "Configuration Error", JOptionPane.WARNING_MESSAGE);
            return;
        }
        Map<String, Mat> outputLayers = getMapOutputLayer(dnnNet);
                
        if (mediaType.equals("Image"))
        {
            detectObjectOnImage(winName, dnnNet, outputLayers, mediaFile, threshold, isSaveOuput, isMaskShading);
        }
        else if (mediaType.equals("Video"))
        {
            System.out.println(outputLayers.size());
            detectObjectOnVideo(winName, dnnNet, outputLayers, mediaFile, threshold, isSaveOuput, isMaskShading);
        }
    }
    
    public void detectObjectOnImage(
        String winName, Net dnnNet, Map<String, Mat> outputLayers, 
        String mediaFile, double threshold, boolean isSaveOuput, boolean isMaskShading
    ) {
       HighGui.namedWindow(winName, HighGui.WINDOW_NORMAL);
       
        Mat frame = Imgcodecs.imread(mediaFile);
        
        int idx = new File(mediaFile).getName().split("\\.").length - 2;
        
        String outputFile 
            = new File(mediaFile).getParent() + File.separator + "out" +  File.separator + 
                new File(mediaFile).getName().split("\\.")[idx] + "_mrcnn_out.jpg";
        
        // Stop the program if reached end of video
        if (frame.empty()) {
            System.out.println( "can not read!!!");
            System.out.println( "Output file is stored as " + outputFile);
            HighGui.waitKey(3000);
            return;
        }
        
        Map<String, Object> resultMap  = resizeMatImage(frame, BLOB_MASK_RCNN_SIZE);
        
        Mat resizedMat = (Mat)resultMap.get("mat");
        Rect padding = (Rect)resultMap.get("padding");
        float scale = (float)resultMap.get("scale");
        
        Mat blob = Dnn.blobFromImage(resizedMat, 1.0, BLOB_MASK_RCNN_SIZE, Scalar.all(0), true, false);
        dnnNet.setInput(blob);
        blob.release();
        
        for(Iterator<String> iter = outputLayers.keySet().iterator(); iter.hasNext();)
        {
            String key = iter.next();
            Mat dtectMat = dnnNet.forward(key);
            outputLayers.put(key, dtectMat);
        }
        
        
        Mat detection = outputLayers.get("detection_out_final");
        Mat detectionMasks = outputLayers.get("detection_masks");
        
        int maxObjCnt = detection.size(2);
        int attrSize = detection.size(3);
        boolean foundObject = false;
        detection = detection.reshape(1, (int) detection.total() / attrSize);
        
//        // 4️⃣ isContinuous 체크 후 clone()
//        if (!detection.isContinuous()) {
//            detection = detection.clone();
//        }
        
        // extract confidence value
        int cols = detection.cols();
        float[] data = new float[(int) (detection.total())];
        detection.get(0, 0, data);  // Mat data to 1D array.
        
        for (int i = 0; i < maxObjCnt; i++) {
            float confidence = data[i * cols + 2];
   
            if (confidence > (threshold <= 0 ? DEFAULT_THRESHOLD : threshold)) {  
                foundObject = true;
                int classId = (int)data[i * cols + 1] ;

                // convert to Bounding Box Conversion
                int left   = (int)((data[i * cols + 3] * MASK_RCNN_SIZE - padding.x) / scale);
                int top    = (int)((data[i * cols + 4] * MASK_RCNN_SIZE - padding.y) / scale);
                int right  = (int)((data[i * cols + 5] * MASK_RCNN_SIZE - padding.x) / scale);
                int bottom = (int)((data[i * cols + 6] * MASK_RCNN_SIZE - padding.y) / scale);
                
                DetectedBBox bbox = new DetectedBBox(confidence, classId, left, top, right, bottom);
                
                
                if(isMaskShading)
                {
                    // 1. Retrieve a mask data of specific class from rowMat to 2D array conversion.
                    Mat rowMat = detectionMasks.row(i).col(bbox.getClassId());

                    // 2. Set the size of mask, gray scale conversion and creation of the object mask.
                    Mat objectMask = new Mat(rowMat.size(2), rowMat.size(3), CvType.CV_32F);
                    rowMat.reshape(1, (int) rowMat.total()).copyTo(objectMask.reshape(1, (int) rowMat.size(2)));
                    drawMaskedBBox(frame, bbox, objectMask, confidence);
                }
                else
                {
                    drawBoxSimple(frame, bbox, confidence);
                }
            }
        }
        
        // Put efficiency information. The function getPerfProfile returns the overall time for inference(t) 
        // and the timings for each of the layers(in layersTimes)
        MatOfDouble layersTimes = new MatOfDouble();
        double freq = Core.getTickFrequency() / 1000;
        double t = dnnNet.getPerfProfile(layersTimes) / freq;
        String label = String.format("Mask-RCNN, Inference time for a frame : %f ms", t);
        Imgproc.putText(frame, label, new Point(0, 15), 
            Imgproc.FONT_HERSHEY_SIMPLEX, 0.5, new Scalar(0, 0, 0));
        
        if (!foundObject) {
            System.out.println("Object Not found.");
        }
        HighGui.imshow(winName, frame);
        if(isSaveOuput)
        {
            // Write the frame with the detection boxes
            Mat detectedFrame = new Mat();
            frame.convertTo(detectedFrame, CvType.CV_8U);

            Imgcodecs.imwrite(outputFile, detectedFrame);
            
            detectedFrame.release();
        }
        //blob.release();
        layersTimes.release();
        
        // Proecess when the window is closing.
        int key = HighGui.waitKey(0); // Wait for key
        if (key == 27) {  // Close when ESC key clicked
            System.out.println("Window closed");
        }

        // Explicit termination of window
        HighGui.destroyAllWindows();  // close all windows
        System.exit(0);  // exit
    }
    
    public void detectObjectOnVideo(
       String winName, Net dnnNet, Map<String, Mat> outputLayers, String mediaFile,
       double threshold, boolean isSaveOuput, boolean isMaskShading
    ) {
       HighGui.namedWindow(winName, HighGui.WINDOW_NORMAL);
        
        Mat frame = new Mat();
        VideoCapture capture = new VideoCapture(mediaFile);
        
        if (capture.open(mediaFile)) {
            capture.read(frame);
            HighGui.imshow(winName, frame);
        } else {
            System.out.println("Capture is not opened!");
            return;
        }
        
        VideoWriter vwriter = new VideoWriter();
        Mat detectedFrame = new Mat();
        String outputFile = mediaFile.substring(0, mediaFile.length() - 4) + "_yolov4_out.avi";
        
        Mat detection;
        Mat detectionMasks;
        
        while (capture.isOpened())
        {
            if(HighGui.waitKey(100) == 27) //ESC key to exit
            {
                System.out.println( "[ESC]Done processing !!!");
                //System.out.println( "[ESC]Output file is stored as " + outputFile);
                HighGui.waitKey(1000);
                break;
            }
            
            // get frame from the video
            capture.read(frame);
           
            // Stop the program if reached end of video
            if (frame.empty()) {
                System.out.println( "Done processing !!!");
                System.out.println( "Output file is stored as " + outputFile);
                HighGui.waitKey(3000);
                 if(vwriter.isOpened()) {
                    vwriter.release();
                } 
                break;
            }

            Map<String, Object> resultMap  = resizeMatImage(frame, BLOB_MASK_RCNN_SIZE);
        
            Mat resizedMat = (Mat)resultMap.get("mat");
            Rect padding = (Rect)resultMap.get("padding");
            float scale = (float)resultMap.get("scale");
            
            // Create a 4D blob from a frame.
            Mat blob = Dnn.blobFromImage(resizedMat, 1.0, BLOB_MASK_RCNN_SIZE, Scalar.all(0), true, false);
            //Sets the input to the network
            dnnNet.setInput(blob);
            
             //List<String> outNames = new ArrayList<>(2);
            for(Iterator<String> iter = outputLayers.keySet().iterator(); iter.hasNext();)
            {
                String key = iter.next();
                //outNames.add(iter.next());
                //System.out.println(key);
                Mat dtectMat = dnnNet.forward(key);
                outputLayers.put(key, dtectMat);
            }
            resizedMat.release();
            blob.release();

            for(Iterator<String> iter = outputLayers.keySet().iterator(); iter.hasNext();)
            {
                String key = iter.next();
                Mat dtectMat = dnnNet.forward(key);
                //System.out.println(key + "detectionMasks size: " + dtectMat.size() + ", rows: " + dtectMat.rows() + ", cols: " + dtectMat.cols());
                outputLayers.put(key, dtectMat);
            }
           
            detection = outputLayers.get("detection_out_final");
            detectionMasks = outputLayers.get("detection_masks");

            int maxObjCnt = detection.size(2);
            int attrSize = detection.size(3);
            boolean foundObject = false;
            detection = detection.reshape(1, (int) detection.total() / attrSize);

            // extract confidence value
            int cols = detection.cols();
            float[] data = new float[(int) (detection.total())];
            detection.get(0, 0, data);  // Mat data to 1D array.
            
            for (int i = 0; i < maxObjCnt; i++) {
                float confidence = data[i * cols + 2];

                if (confidence > (threshold <= 0 ? DEFAULT_THRESHOLD : threshold)) {  
                    //System.out.println("threshold >>>>" + threshold);
                    int classId = (int)data[i * cols + 1] ;

                    // convert to Bounding Box Conversion
                    int left   = (int)((data[i * cols + 3] * MASK_RCNN_SIZE - padding.x) / scale);
                    int top    = (int)((data[i * cols + 4] * MASK_RCNN_SIZE - padding.y) / scale);
                    int right  = (int)((data[i * cols + 5] * MASK_RCNN_SIZE - padding.x) / scale);
                    int bottom = (int)((data[i * cols + 6] * MASK_RCNN_SIZE - padding.y) / scale);

                    DetectedBBox bbox = new DetectedBBox(confidence, classId, left, top, right, bottom);

                    if(isMaskShading)
                    {
                         // 1. Retrieve a mask data of specific class from rowMat to 2D array conversion.
                        Mat rowMat = detectionMasks.row(i).col(bbox.getClassId());

                        // 2. Set the size of mask, gray scale conversion and creation of the object mask.
                        Mat objectMask = new Mat(rowMat.size(2), rowMat.size(3), CvType.CV_32F);
                        rowMat.reshape(1, (int) rowMat.total()).copyTo(objectMask.reshape(1, (int) rowMat.size(2)));
                    
                        drawMaskedBBox(frame, bbox, objectMask,confidence);
                        
                        objectMask.release();
                        rowMat.release();
                    }
                    else
                    {
                        drawBoxSimple(frame, bbox,confidence);
                    }
                }
                
                detection.release();
            }

            // Put efficiency information. The function getPerfProfile returns the overall time for inference(t) 
            //and the timings for each of the layers(in layersTimes)
            MatOfDouble layersTimes = new MatOfDouble();
            double freq = Core.getTickFrequency() / 1000;
            double t = dnnNet.getPerfProfile(layersTimes) / freq;
            String label = String.format("%s, Inference time for a frame : %f ms", modelName, t);
            Imgproc.putText(frame, label, new Point(0, 15), 
                            Imgproc.FONT_HERSHEY_SIMPLEX, 0.5, new Scalar(0, 0, 0), 2);
            
            HighGui.imshow(winName, frame);

            if(isSaveOuput)
            {
                double fps = capture.get(Videoio.CAP_PROP_FPS);
                System.out.println("fps: " + fps);
                if(fps == 0D)
                {
                    //fps = 20;
                    continue;
                }
                
                Size size 
                    = new Size(capture.get(Videoio.CAP_PROP_FRAME_WIDTH), capture.get(Videoio.CAP_PROP_FRAME_HEIGHT));

    //            int fourcc_xvid = VideoWriter.fourcc('X', 'V', 'I', 'D');   //# XVID (AVI format)
    //            int fourcc_mp4v = VideoWriter.fourcc('M', 'P', '4', 'V');  // # MP4 format
                int fourcc_mjpg = VideoWriter.fourcc('M', 'J', 'P', 'G');  //# Motion JPEG format
    //            int fourcc_h264 = VideoWriter.fourcc('H', '2', '6', '4'); // # H264 (more advanced compression)

                isSaveOuput = vwriter.open(outputFile, fourcc_mjpg, fps, size, true);

                if(!isSaveOuput)
                {
                    JOptionPane.showMessageDialog(null, 
                        String.format("Can not open the video file[sz:%s, fps:%f]", size.toString(), fps), 
                        "error", JOptionPane.INFORMATION_MESSAGE);
                    break;
                }
                
                frame.convertTo(detectedFrame, CvType.CV_8U);
                if(vwriter.isOpened()) 
                {
                    vwriter.write(detectedFrame);
                }
                else
                {
                    System.out.println("========vwriter is not opened=======");
                }
            }

            layersTimes.release();
            
            // Proecess when the window is closing.
//            int key = HighGui.waitKey(0); // Wait for key
//            if (key == 27) {  // Close when ESC key clicked
//                System.out.println("Window closed");
//            }
        }
        
        
        // Explicit termination of window
        HighGui.destroyAllWindows();  // close all windows
        System.exit(0);  // exit
    }
    
    // Draw the predicted bounding box, colorize and show the mask on the image
    private boolean drawMaskedBBox(
        Mat frame, DetectedBBox bbox, Mat objectMask,float confidence
    ) {
        boolean foundObject = true;
        // create an object for label information.
        Map<Integer, String> cocoLabels =  CocoLabel.getCocoLabel();
        // generate radnom color in order to draw bounding boxes
        List<Scalar> colors = CocoLabel.getColorsByLabel();
        String label = String.format("%s: %.2f", cocoLabels.get(bbox.getClassId()), confidence * 100);
        Scalar classColor = colors.get(bbox.getClassId());
        
        Rect box = new Rect(bbox.getLeft(), bbox.getTop(), bbox.getRight() - bbox.getLeft() + 1, bbox.getBottom() - bbox.getTop() + 1);
        
        // Display the label at the top of the bounding box
        int[] baseLine = { 0 };
        Size labelSize = Imgproc.getTextSize(label, Imgproc.FONT_HERSHEY_SIMPLEX, 0.5, 1, baseLine);
        box.y = Math.max(box.y, (int) labelSize.height);
        // ✅ Drawing a Bounding Box Only
        //Imgproc.rectangle(frame, new Point(box.x, box.y), new Point(bbox.right, bbox.bottom), classColor, 2);
        // Draw background for text
        Imgproc.rectangle(
            frame, new Point(box.x, box.y - Math.round(1.5 * labelSize.height)),
            new Point(box.x + Math.round(1.5 * labelSize.width), box.y + baseLine[0]), 
            new Scalar(255, 255, 255), Imgproc.FILLED);

        Imgproc.putText(frame, label, new Point(bbox.getLeft(), bbox.getTop() - 5), Imgproc.FONT_HERSHEY_SIMPLEX, 0.5, classColor, 2);
        
        // frame이 비어있는지 확인
        if (frame.empty()) {
            System.out.println("frame is empty! failed to load an image.");
            return false;
        }

        // check the frame vaildation
        int width = frame.cols();
        int height = frame.rows();
        if (width <= 0 || height <= 0) {
            System.out.println("invalid frame size.");
            return false;
        }
        
        // 3. Set the threshold for adjusting  binarizatiion of the mask 
        Imgproc.threshold(objectMask, objectMask, 0.5, 1.0, Imgproc.THRESH_BINARY);

        // 4. About to fuilfil color by Converting `objectMask` to `CV_8UC1`
        objectMask.convertTo(objectMask, CvType.CV_8UC1); // 변환

        // 5. Shading Color making a gray scale `objectMask` 
        Mat colorMask = new Mat();
        Imgproc.cvtColor(objectMask, colorMask, Imgproc.COLOR_GRAY2BGR);

        // 6. Apply the color mask :  classColor
        Core.multiply(colorMask, classColor, colorMask); // Adjust masks mutiply by the color

        // 7. Smoothing the edges
        Mat smoothMask = new Mat();
        Imgproc.GaussianBlur(colorMask, smoothMask, new Size(objectMask.rows(), objectMask.cols()), 0);
        
        // 8. Apply the ROI.
        //Mat roi = frame.submat(bbox.getTop(), bbox.getBottom(), bbox.getLeft(), bbox.getRight());
        Rect box1 = new Rect(bbox.getLeft(), bbox.getTop(), bbox.getRight() - bbox.getLeft() + 1, bbox.getBottom() - bbox.getTop() + 1);;
        Mat roi = getSafeROI(frame, box1);
        // 9. Check the same size of both the colcoMask and roi, if not resize the image.
        if (roi.size().width != smoothMask.size().width || roi.size().height != smoothMask.size().height) {
            Imgproc.resize(smoothMask, smoothMask, roi.size());
        }

        // 10. Convert the data type of `smoothMask` to `CV_8U`covered by `roi`
        smoothMask.convertTo(smoothMask, CvType.CV_8U);

        // 11. Draw a BBox (adding a boundary for the original image)
        Imgproc.rectangle(frame, new Point(bbox.getLeft(), bbox.getTop()), new Point(bbox.getRight(), bbox.getBottom()), classColor, 2); // BBox

        // 12. Paint the mask to the original image(Alpha Blending)
        Core.addWeighted(roi, 0.8, smoothMask, 0.5, 0, roi);
        
        return foundObject;
    }
    
    // Draw the predicted bounding box, colorize and show the mask on the image
    private static void drawBoxSimple(
        Mat frame, DetectedBBox bbox, float confidence
    ) {
        // create an object for label information.
        Map<Integer, String> cocoLabels =  CocoLabel.getCocoLabel();
        // generate radnom color in order to draw bounding boxes
        List<Scalar> colors = CocoLabel.getColorsByLabel();
        String label = String.format("%s: %.2f", cocoLabels.get(bbox.getClassId()), confidence * 100);
        Scalar classColor = colors.get(bbox.getClassId());
        
        Rect box = new Rect(bbox.getLeft(), bbox.getTop(), bbox.getRight() - bbox.getLeft() + 1, bbox.getBottom() - bbox.getTop() + 1);
        //Scalar selectedColor = colors.get(classId);
        Point x_y = new Point(box.x, box.y);
        Point w_h = new Point(box.x + box.width, box.y + box.height);
        //Point text_point = new Point(box.x, box.y - 5);
        //String label = cocoLabels.get(classId);
        
        Imgproc.rectangle(frame, w_h, x_y, classColor, 1);
        
        int baseLine[] = { 0 };
        Size labelSize = Imgproc.getTextSize(label, Imgproc.FONT_HERSHEY_SIMPLEX, 0.5, 1, baseLine);
        box.y = Integer.max(box.y, (int) labelSize.height);
        
        // ✅ Drawing a Bounding Box Only
        //Imgproc.rectangle(frame, new Point(box.x, box.y), new Point(bbox.right, bbox.bottom), classColor, 2);
        // Draw background for text
        Imgproc.rectangle(
            frame, new Point(box.x, box.y - Math.round(1.5 * labelSize.height)),
            new Point(box.x + Math.round(1.5 * labelSize.width), box.y + baseLine[0]), 
            new Scalar(255, 255, 255), Imgproc.FILLED);
        
        Imgproc.putText(frame, label, new Point(box.x, box.y), 
            Imgproc.FONT_HERSHEY_SIMPLEX, 0.5, classColor,1);
    }
}
