package tobee.test;

import com.tobee.opencv.detector.tests.CvUtils;
import org.opencv.core.*;
import org.opencv.dnn.*;
import org.opencv.imgcodecs.*;
import org.opencv.imgproc.*;
import org.opencv.core.CvType;

import java.util.ArrayList;
import java.util.List;
import org.opencv.highgui.HighGui;

public class MaskRCNNTest3 {
    public static final float CONFIDENCE_THRESHOLD = 0.5f; // Confidence threshold for filtering detections

    public static void postprocess(Mat frame, List<Mat> outs) {
        // Output of the network
        Mat outDetections = outs.get(0);
        Mat outMasks = outs.get(1);

        // Output size of masks is NxCxHxW where:
        // N - number of detected boxes
        // C - number of classes (excluding background)
        // HxW - segmentation shape
        int numDetections = outDetections.size(2);
        int numClasses = outMasks.size(1);

        // Reshape outDetections to (N x 7)
        outDetections = outDetections.reshape(1, (int) outDetections.total() / 7);

        for (int i = 0; i < numDetections; ++i) {
            float score = (float) outDetections.get(i, 2)[0];
            if (score > CONFIDENCE_THRESHOLD) {
                // Extract the bounding box
                int classId = (int) outDetections.get(i, 1)[0];
                int left = (int) (frame.cols() * outDetections.get(i, 3)[0]);
                int top = (int) (frame.rows() * outDetections.get(i, 4)[0]);
                int right = (int) (frame.cols() * outDetections.get(i, 5)[0]);
                int bottom = (int) (frame.rows() * outDetections.get(i, 6)[0]);

                // Ensure bounding box coordinates are within image boundaries
                left = Math.max(0, Math.min(left, frame.cols() - 1));
                top = Math.max(0, Math.min(top, frame.rows() - 1));
                right = Math.max(0, Math.min(right, frame.cols() - 1));
                bottom = Math.max(0, Math.min(bottom, frame.rows() - 1));

                Rect box = new Rect(left, top, right - left + 1, bottom - top + 1);

                // Extract the mask for the object
                //Mat objectMask = outMasks.row(i).row(classId).reshape(1, outMasks.size(2)); // Get mask for object
                //Mat objectMask = new Mat(outMasks.size(2), outMasks.size[3],CV_32F, outMasks.ptr<float>(i,classId));
                float fVal = (float)CvUtils.getMatVal(outDetections, i, 1);
                Mat objectMask = new Mat(outMasks.size(2), outMasks.size(3), CvType.CV_32F, new Scalar(fVal));
                // Resize the mask to match the original image size
                Mat resizedMask = new Mat();
                Imgproc.resize(objectMask, resizedMask, new Size(frame.cols(), frame.rows()));

                // Draw bounding box, colorize and show the mask on the image
                drawBox(frame, classId, score, box, resizedMask);
            }
        }
    }

    private static void drawBox(Mat frame, int classId, float score, Rect box, Mat mask) {
        // Optionally, you can draw the bounding box with a different color for each class
        Scalar color = new Scalar(0, 255, 0);  // Green color for box

        // Draw the bounding box
        Imgproc.rectangle(frame, box.tl(), box.br(), color, 2);

        // Optionally, you can show the mask by blending it with the image
        // Here, you can use a colorized mask overlay with the bounding box
        for (int y = box.y; y < box.y + box.height; y++) {
            for (int x = box.x; x < box.x + box.width; x++) {
                if (mask.get(y, x)[0] > 0.5) { // Mask threshold to check if pixel is part of the mask
                    frame.put(y, x, color.val); // Colorize the pixel as part of the mask
                }
            }
        }

        // Optionally, draw the class label and score
        String label = "Class " + classId + ": " + score;
        Imgproc.putText(frame, label, box.tl(), Imgproc.FONT_HERSHEY_SIMPLEX, 0.5, color, 1);
    }
    public static void main(String[] args) {
        // OpenCV 라이브러리 로드
        System.loadLibrary(Core.NATIVE_LIBRARY_NAME);

        // 모델 파일 경로 설정
        String modelPath = "C:\\DEV\\Experimental\\Vision\\pretrained_model\\mrcnn\\frozen_inference_graph.pb";
        String configPath = "C:\\DEV\\Experimental\\Vision\\pretrained_model\\mrcnn\\mask_rcnn_inception_v2_coco_2018_01_28.pbtxt";

        // 모델 로드
        Net net = Dnn.readNetFromTensorflow(modelPath, configPath);

        // 이미지 로드
        Mat orgMat = Imgcodecs.imread("C:\\DEV\\Experimental\\Vision\\data\\images\\2516944023_d00345997d_z.jpg");

        if (orgMat.empty()) {
            System.out.println("❌ 이미지 로드 실패!");
            return;
        }

        // 크기 설정
        int targetWidth = 800;
        int targetHeight = 800;

        // 리사이즈 (letterbox 방식)
        Mat calcMat = letterboxResize(orgMat, targetWidth, targetHeight);

        // Blob 생성 (mean subtraction과 RGB->BGR 변환)
        Mat blob = Dnn.blobFromImage(orgMat, 1.0, new Size(targetWidth, targetHeight),
                                     new Scalar(102.9801, 115.9465, 122.7717), true, false);
        net.setInput(blob);

        // 모델 출력 레이어
        List<String> modelOutputs = new ArrayList<>();
        modelOutputs.add("detection_out_final");
        modelOutputs.add("detection_masks");

        // 출력 받기
        List<Mat> outputs = new ArrayList<>();
        net.forward(outputs, modelOutputs);
        
        postprocess(orgMat, outputs);
        
        // 결과 출력
        HighGui.imshow("Result", orgMat);
        HighGui.waitKey(0);
    }

    // Letterbox 리사이징 함수
    public static Mat letterboxResize(Mat orgMat, int targetWidth, int targetHeight) {
        int originalWidth = orgMat.cols();
        int originalHeight = orgMat.rows();

        // 비율 유지한 채 리사이징할 크기 계산
        float scaleX = (float) targetWidth / originalWidth;
        float scaleY = (float) targetHeight / originalHeight;
        float minScale = Math.min(scaleX, scaleY);

        int newWidth = (int) (originalWidth * minScale);
        int newHeight = (int) (originalHeight * minScale);

        // 패딩 값 계산
        int left = (targetWidth - newWidth) / 2;
        int top = (targetHeight - newHeight) / 2;
        int right = targetWidth - newWidth - left;
        int bottom = targetHeight - newHeight - top;

        // 리사이징 및 패딩 적용
        Mat resizedMat = new Mat();
        Imgproc.resize(orgMat, resizedMat, new Size(newWidth, newHeight));

        Mat paddedMat = new Mat();
        Core.copyMakeBorder(resizedMat, paddedMat, top, bottom, left, right, Core.BORDER_CONSTANT, new Scalar(0, 0, 0));

        return paddedMat;
    }
}

   